'use strict';

var tslib = require('tslib');
var uuid = require('uuid');
var React = require('react');
var imageDisplayControlMetadataParser = require('@frameright/image-display-control-metadata-parser');

function _interopNamespaceDefault(e) {
    var n = Object.create(null);
    if (e) {
        Object.keys(e).forEach(function (k) {
            if (k !== 'default') {
                var d = Object.getOwnPropertyDescriptor(e, k);
                Object.defineProperty(n, k, d.get ? d : {
                    enumerable: true,
                    get: function () { return e[k]; }
                });
            }
        });
    }
    n.default = e;
    return Object.freeze(n);
}

var React__namespace = /*#__PURE__*/_interopNamespaceDefault(React);

// If true, we are either running on the server at request-time, or at
// build time, building a static (e.g. Next.js) page.
const isServerOrStatic = typeof window === 'undefined';
const isBrowser = !isServerOrStatic;
let isRunningTests = false;
try {
    // See https://stackoverflow.com/questions/50940640/how-to-determine-if-jest-is-running-the-code-or-not
    isRunningTests = process.env.JEST_WORKER_ID !== undefined;
}
catch (e) {
    // Probably "ReferenceError: process is not defined", ignore.
}
// See https://docs.frameright.io/web-component/importing
if (isBrowser && !isRunningTests) {
    // Defines the <img is="image-display-control"> web component.
    import('@frameright/image-display-control-web-component/dist/src/image-display-control.js');
}
let fs = null;
if (!isBrowser) {
    // We need to use `require()` here with Vite/Vike in prod as otherwise
    // `import()` will happen later than the server-side rendering. However
    // `require()` isn't defined in dev mode, so then we fall back to `import()`.
    if (require) {
        // FIXME: for this to work on Next.js, we need to set
        // `config.resolve.fallback = { fs: false };` in next.config.js. See
        // https://stackoverflow.com/questions/64926174/module-not-found-cant-resolve-fs-in-next-js-application
        fs = require('fs');
    }
    else {
        import('fs').then((fsModule) => {
            fs = fsModule;
        });
    }
}
function ImageDisplayControl({ children, 'data-debug': debug = false, }) {
    _traceIfDebug(debug, 'Rendering...');
    const allChildren = children || [];
    // Same UUID and ref for all <img>-like children of one same
    // <ImageDisplayControl> component.
    //
    // Note: it may happen that when adding new children later, they will obtain
    // a different UUID than previous children. This is fine, as this UUID is only
    // used for finding server-side rendered image regions in the DOM on initial
    // hydration.
    // New children will not have any new server-side rendered image regions to
    // take from the DOM. Instead, all initially server-side rendered image
    // regions will be already present in the state anyway.
    const [uuid$1] = React__namespace.useState(() => uuid.v4());
    const ref = React__namespace.useRef(null);
    // This maps an image src to its image regions as a json string. Let's
    // initially render the web component with empty image regions. Image regions
    // will be fetched asynchronously with useEffect().
    const [imageRegionsMap, setImageRegionsMap] = React__namespace.useState(new Map());
    _initializeImageRegionsMap(imageRegionsMap, allChildren, debug);
    // We are going to extend all <img>-like elements by adding attributes to
    // them, in order to turn them into <img is="image-display-control"> elements
    // under the hood.
    let numImgChildren = 0;
    const extendedChildren = React__namespace.Children.map(allChildren, (child) => {
        const imageSource = _getImageSource(child);
        if (!imageSource) {
            // This child isn't an <img>-like element. No need to extend it with the
            // web component functionality.
            return child;
        }
        ++numImgChildren;
        const newAttrs = {
            // Note: thanks to this, react will recognize this as a custom element.
            // https://react.dev/reference/react-dom/components#custom-html-elements
            is: 'image-display-control',
            // Because this is a custom element, `class=` should be used instead of
            // `className=`. However we want to allow the user to use `className=`, so
            // we copy it over to `class=`:
            class: `${child.props.class || ''} ${child.props.className || ''}`.trim(),
            ref,
            'data-idc-uuid': uuid$1,
            // There might be a `data-image-regions=` attribute mismatch between
            // what's rendered on the server and what's rendered during the initial
            // client-side hydration. This is because for performance reasons, if the
            // server has already parsed the image regions, we don't want to do this
            // work again in the browse. Instead we don't explicitly re-render the
            // `data-image-regions=` attribute, see below. This makes React issue a
            // warning, which we suppress here. See also
            // * https://react.dev/reference/react-dom/client/hydrateRoot#suppressing-unavoidable-hydration-mismatch-errors
            // * https://stackoverflow.com/questions/40093655/how-do-i-add-attributes-to-existing-html-elements-in-typescript-jsx
            suppressHydrationWarning: true,
            // On Next.js in prod, the react `src` prop may end up being different
            // from the rendered DOM `src` attribute. Keep track of the prop for
            // mapping.
            'data-src-prop': imageSource.src,
        };
        if (isServerOrStatic) {
            // We don't want to leak `data-path-on-server` to the client, so we remove
            // it from the attributes.
            newAttrs['data-path-on-server'] = undefined;
        }
        const knownImageRegions = imageRegionsMap.get(imageSource.src);
        if (knownImageRegions) {
            newAttrs['data-image-regions'] = knownImageRegions;
        }
        const extendedChild = React__namespace.cloneElement(child, newAttrs);
        _traceIfDebug(debug, 'Rendering extended child:', extendedChild);
        return extendedChild;
    });
    // See
    // https://stackoverflow.com/questions/57847626/using-async-await-inside-a-react-functional-component
    React__namespace.useEffect(() => {
        _populateImageRegionsMap(imageRegionsMap, setImageRegionsMap, ref, debug);
        _checkParentElement(ref);
    }, [imageRegionsMap, ref, debug, numImgChildren]);
    return React__namespace.createElement(React__namespace.Fragment, null, extendedChildren);
}
// This function makes sure that each <img>-like child has an entry in the map,
// worst case with no image regions.
// If we are on server-side, it tries to synchronously reads the image regions
// from the metadata of the image file on disk.
function _initializeImageRegionsMap(imageRegionsMap, children, debug) {
    _traceIfDebug(debug, 'Initializing image regions map...');
    React__namespace.Children.forEach(children, (child) => {
        _traceIfDebug(debug, 'Initializing image regions map for child:', child);
        const imageSource = _getImageSource(child);
        if (!imageSource) {
            // This child isn't an <img>-like element, skipping.
            return;
        }
        if (!imageRegionsMap.get(imageSource.src)) {
            let imageRegions = '';
            if (isServerOrStatic) {
                if (!imageSource.pathOnServer) {
                    _warn('Missing data-path-on-server attribute for', imageSource.src, ", can't read image regions from disk. You probably want to read " +
                        'https://docs.frameright.io/react/ssr');
                }
                else if (imageSource.pathOnServer !== 'none') {
                    _traceIfDebug(debug, 'Reading image regions on disk for', imageSource.src);
                    imageRegions = _readImageRegionsJsonFromDisk(imageSource.pathOnServer, debug);
                }
            }
            imageRegionsMap.set(imageSource.src, imageRegions);
        }
        else {
            _traceIfDebug(debug, 'Found image regions in state for', imageSource.src);
        }
    });
}
// On client-side, for each key of the map, which is an image src, fetch/parse
// the image regions from the image's metadata, if it hasn't been done already,
// and populate the map accordingly.
// Not only it populates the map, but it also calls the provided react state
// setter.
//
// Note: imgChildRef is a ref to one of the <img>-like children.
function _populateImageRegionsMap(imageRegionsMap, setImageRegionsMap, imgChildRef, debug) {
    return tslib.__awaiter(this, void 0, void 0, function* () {
        _traceIfDebug(debug, 'Populating image regions map if needed...');
        // Building a map of image regions from the pre-rendered DOM, if any.
        const imageRegionsMapFromDom = new Map();
        for (const child of _getImgChildren(imgChildRef)) {
            const imageSource = _getImageSource(child);
            if (imageSource) {
                _traceIfDebug(debug, 'Reading image regions from DOM for', imageSource.src, '...');
                const imageRegions = child.dataset.imageRegions;
                if (imageRegions) {
                    imageRegionsMapFromDom.set(imageSource.src, imageRegions);
                }
            }
        }
        let modified = false;
        for (const [src, imageRegions] of imageRegionsMap) {
            if (!imageRegions) {
                const imageRegionsFromDom = imageRegionsMapFromDom.get(src);
                if (imageRegionsFromDom) {
                    // No need to fetch/parse the image regions from the image's metadata,
                    // we already have them from the pre-rendered DOM.
                    _traceIfDebug(debug, 'Populating image regions from DOM for', src);
                    imageRegionsMap.set(src, imageRegionsFromDom);
                }
                else {
                    // The regions for this image haven't been fetched yet. Let's fetch them
                    // now.
                    _traceIfDebug(debug, 'Image regions not found in DOM for', src, ', fetching...');
                    let arrayBuffer;
                    try {
                        // Notes:
                        // * may or may not be in cache yet.
                        // * may lead to CORS-related errors.
                        const image = yield fetch(src);
                        arrayBuffer = yield image.arrayBuffer();
                    }
                    catch (error) {
                        _warn('Error fetching image:', error);
                        continue;
                    }
                    // Because of the awaits, the regions may have been populated already by
                    // another promise. Let's check again:
                    if (imageRegionsMap.get(src)) {
                        _traceIfDebug(debug, 'Image regions for', src, 'already populated meanwhile, aborting.');
                        continue;
                    }
                    yield _waitForImports(debug);
                    const regions = _readImageRegionsJsonFromBuffer(arrayBuffer);
                    if (regions) {
                        _traceIfDebug(debug, 'Populating image regions from metadata for', src);
                        imageRegionsMap.set(src, regions);
                        modified = true;
                    }
                }
            }
        }
        if (modified) {
            _traceIfDebug(debug, 'Updating image regions map in state.');
            setImageRegionsMap(new Map(imageRegionsMap));
        }
        else {
            _traceIfDebug(debug, 'Image regions map in state is up-to-date.');
        }
        _traceIfDebug(debug, 'Image regions map in state:', imageRegionsMap);
    });
}
// Note: imgChildRef is a ref to one of the <img>-like children.
function _checkParentElement(imgChildRef) {
    const child = imgChildRef.current;
    if (child) {
        const parent = child.parentElement;
        if (parent && !parent.dataset.idcParent) {
            _warn("Parent element of <ImageDisplayControl> doesn't have a " +
                "'data-idc-parent' attribute. You probably want to read " +
                'https://docs.frameright.io/react/attributes#parent-dom-elements-attributes');
        }
    }
}
// Note: imgChildRef is a ref to one of the <img>-like children.
function _getImgChildren(imgChildRef) {
    const child = imgChildRef.current;
    if (child) {
        // Note: we use the UUID to make sure we don't catch neighboring images
        // managed by another <ImageDisplayControl>:
        //
        // <div> <-- DOM parent of all the <img>s
        //   <ImageDisplayControl> <-- doesn't produce a DOM element
        //     <img data-idc-uuid="123" />
        //     <img data-idc-uuid="123" />
        //   </ImageDisplayControl>
        //   <ImageDisplayControl> <-- doesn't produce a DOM element
        //     <img data-idc-uuid="456" /> <-- DOM sibling of the previous <img>s
        //   </ImageDisplayControl>
        // </div>
        const uuid = child.dataset.idcUuid;
        if (uuid) {
            const parent = child.parentElement;
            if (parent) {
                return Array.from(parent.querySelectorAll('[data-idc-uuid="' + uuid + '"]'));
            }
        }
    }
    return [];
}
// To be used only on server-side.
function _readImageRegionsJsonFromDisk(pathOnServer, debug) {
    if (!fs) {
        _warn("fs module not available, can't read image regions from disk.");
        return '';
    }
    let result = '';
    try {
        _traceIfDebug(debug, 'Reading image regions from', pathOnServer);
        const buffer = fs.readFileSync(pathOnServer);
        result = _readImageRegionsJsonFromBuffer(buffer);
        _traceIfDebug(debug, 'Found image regions:', result);
    }
    catch (error) {
        _warn('Error while reading image regions from disk:', error);
    }
    return result;
}
function _readImageRegionsJsonFromBuffer(buffer) {
    const parser = new imageDisplayControlMetadataParser.Parser(buffer);
    const regions = parser.getIdcMetadata('rectangle', 'crop');
    return JSON.stringify(regions);
}
function _getImageSource(element) {
    var _a, _b;
    if (!_isImgElement(element)) {
        return null;
    }
    let result = null;
    if ('props' in element) {
        // React element
        if (typeof element.props.src === 'string') {
            result = {
                src: element.props.src,
            };
        }
        else if ('src' in element.props.src) {
            // This typically happen when statically importing an image on Next.js,
            // the resulting object being an instance of StaticImageData.
            result = {
                src: element.props.src.src,
            };
            if ('pathOnServer' in element.props.src) {
                result.pathOnServer = element.props.src.pathOnServer;
            }
        }
        else {
            _warn('Unknown src= attribute format: ', element.props.src);
            return null;
        }
        if (element.props['data-path-on-server']) {
            result.pathOnServer = element.props['data-path-on-server'];
        }
    }
    else {
        // HTML element
        result = {
            src: ((_a = element.attributes.getNamedItem('data-src-prop')) === null || _a === void 0 ? void 0 : _a.value) ||
                ((_b = element.attributes.getNamedItem('src')) === null || _b === void 0 ? void 0 : _b.value) ||
                element.src,
        };
    }
    return result;
}
// Returns true if the element is an <img>-like element, i.e. also Next.js
// <Image> or react-bootstrap <Image> elements, for example.
function _isImgElement(element) {
    var _a;
    if ('props' in element) {
        // React element
        return !!element.props.src;
    }
    else if ('attributes' in element) {
        // HTML element
        return !!((_a = element.attributes.getNamedItem('src')) === null || _a === void 0 ? void 0 : _a.value);
    }
    return false;
}
// FIXME: workaround for async import() on Vite/Vike in dev mode.
function _waitForImports(debug) {
    return tslib.__awaiter(this, void 0, void 0, function* () {
        for (const waitMs of [100, 200, 500, 1000, 1000]) {
            if (fs || isBrowser) {
                _traceIfDebug(debug, 'All imports have resolved.');
                return;
            }
            _traceIfDebug(debug, 'Waiting for imports to resolve...');
            yield new Promise((resolve) => setTimeout(resolve, waitMs));
        }
        _warn('Some imports are still missing, image regions will not be available.');
    });
}
const consolePrefix = '[idc]';
function _warn(...args) {
    console.warn(consolePrefix, '[warn]', ...args);
}
function _traceIfDebug(debug, ...args) {
    if (debug) {
        console.log(consolePrefix, '[debug]', ...args);
    }
}

exports.ImageDisplayControl = ImageDisplayControl;
//# sourceMappingURL=react-image-display-control.cjs.map
